#include <stdio.h>
#include <string.h>

#include <scheduler.h>
#include <timer.h>

#include <aversive/wait.h>
#include <uart.h>
#include <i2c.h>

#include "i2c_helper.h"

#include "mpu6050.h"

#define ToRad(x) ((x)*0.01745329252)  // *pi/180

// fs_sel = 1
// 500°/s
#define gyro_x_resolution 66.5
#define gyro_y_resolution 66.5
#define gyro_z_resolution 66.5

// fs_sel = 3
// 2000°/s
#define gyro_x_resolution 16.4
#define gyro_y_resolution 16.4
#define gyro_z_resolution 16.4

#define accel_x_resolution 8192.0
#define accel_y_resolution 8192.0
#define accel_z_resolution 8192.0




#define MPU6050_OFF_MSB 1
#define MPU6050_OFF_LSB 0


#define MPU6050_ADDRESS 0x68
#define MPU6050_MAGNETO_ADDRESS 0x0C

#define SWAP_16(a) ((( (a) & 0xff)<<8) | (( (a) >> 8) & 0xff))

double mpu6050_gx;
double mpu6050_gy;
double mpu6050_gz;

double mpu6050_ax;
double mpu6050_ay;
double mpu6050_az;

double mpu6050_mx;
double mpu6050_my;
double mpu6050_mz;

double mpu6050_temp;

int16_t drift_g[3] = {0, 0, 0};


uint8_t mpu6050_read_gyro_raw(int16_t *values)
{
	uint8_t err;
	uint8_t i;
	err = read_reg_len(MPU6050_ADDRESS, 0x43, (uint8_t*)values, sizeof(int16_t)*3);
	for (i=0;i<3; i++) {
		values[i] = SWAP_16(values[i]);
	}

	return err;
}

uint8_t mpu6050_read_all_axes(int16_t *values)
{
	uint8_t err;
	uint8_t i;
	err = read_reg_len(MPU6050_ADDRESS, 0x3b, (uint8_t*)values, sizeof(int16_t)*10);

	for (i=0;i<7; i++) {
		values[i] = SWAP_16(values[i]);
	}

	mpu6050_ax = 9.81 * (double)values[0] / accel_x_resolution ;
	mpu6050_ay = 9.81 * (double)values[1] / accel_y_resolution ;
	mpu6050_az = 9.81 * (double)values[2] / accel_z_resolution ;

	mpu6050_temp = (double)values[3]/340. + 36.5;

	mpu6050_gx = ToRad((double)(values[4] - drift_g[0]) / gyro_x_resolution );
	mpu6050_gy = ToRad((double)(values[5] - drift_g[1]) / gyro_y_resolution );
	mpu6050_gz = ToRad((double)(values[6] - drift_g[2]) / gyro_z_resolution );

	mpu6050_mx = (double) values[7] * 0.3;
	mpu6050_my = (double) values[8] * 0.3;
	mpu6050_mz = (double) values[9] * 0.3;


	return err;
}





uint8_t send_mpu6050_cmd(uint8_t address, uint8_t reg, uint8_t val)
{
	uint8_t err;
	uint8_t buffer[2];

	buffer[0] = reg;
	buffer[1] = val;

	err = i2c_send(address, (unsigned char*)buffer, 2, I2C_CTRL_SYNC);
	if (err)
		printf("error %.2X\r\n", err);
	return err;
}


void mpu6050_compute_drift(void)
{
	uint16_t i;
	int32_t s_gx, s_gy, s_gz;
	int16_t g_values[3];

	drift_g[0] = 0;
	drift_g[1] = 0;
	drift_g[2] = 0;

	s_gx = s_gy = s_gz = 0;

	for (i = 0; i < 0x100; i ++) {
		mpu6050_read_gyro_raw(&g_values);
		s_gx += g_values[0];
		s_gy += g_values[1];
		s_gz += g_values[2];
	}
	printf("%"PRId32" %"PRId32" %"PRId32" (%"PRId32") \r\n", s_gx, s_gy, s_gz, i);
	s_gx /= i;
	s_gy /= i;
	s_gz /= i;

	drift_g[0] = s_gx;
	drift_g[1] = s_gy;
	drift_g[2] = s_gz;
	printf("gyro drift:\r\n");
	printf("%d %d %d\r\n",
	       drift_g[0],
	       drift_g[1],
	       drift_g[2]);
}


uint8_t setup_mpu9150_magneto()
{
	//MPU9150_I2C_ADDRESS = 0x0C;      //change Adress to Compass

	send_mpu6050_cmd(MPU6050_MAGNETO_ADDRESS, 0x0A, 0x00); //PowerDownMode
	send_mpu6050_cmd(MPU6050_MAGNETO_ADDRESS, 0x0A, 0x0F); //SelfTest
	send_mpu6050_cmd(MPU6050_MAGNETO_ADDRESS, 0x0A, 0x00); //PowerDownMode

	//MPU9150_I2C_ADDRESS = 0x69;      //change Adress to MPU

	send_mpu6050_cmd(MPU6050_ADDRESS, 0x24, 0x40); //Wait for Data at Slave0
	send_mpu6050_cmd(MPU6050_ADDRESS, 0x25, 0x8C); //Set i2c address at slave0 at 0x0C
	//send_mpu6050_cmd(MPU6050_ADDRESS, 0x26, 0x02); //Set where reading at slave 0 starts
	send_mpu6050_cmd(MPU6050_ADDRESS, 0x26, 0x03); //Set where reading at slave 0 starts
	//send_mpu6050_cmd(MPU6050_ADDRESS, 0x27, 0x88); //set offset at start reading and enable
	send_mpu6050_cmd(MPU6050_ADDRESS, 0x27, 0x86); //set offset at start reading and enable
	send_mpu6050_cmd(MPU6050_ADDRESS, 0x28, 0x0C); //set i2c address at slv1 at 0x0C
	send_mpu6050_cmd(MPU6050_ADDRESS, 0x29, 0x0A); //Set where reading at slave 1 starts
	send_mpu6050_cmd(MPU6050_ADDRESS, 0x2A, 0x81); //Enable at set length to 1
	send_mpu6050_cmd(MPU6050_ADDRESS, 0x64, 0x01); //overvride register
	send_mpu6050_cmd(MPU6050_ADDRESS, 0x67, 0x03); //set delay rate
	send_mpu6050_cmd(MPU6050_ADDRESS, 0x01, 0x80);

	send_mpu6050_cmd(MPU6050_ADDRESS, 0x34, 0x04); //set i2c slv4 delay
	send_mpu6050_cmd(MPU6050_ADDRESS, 0x64, 0x00); //override register
	send_mpu6050_cmd(MPU6050_ADDRESS, 0x6A, 0x00); //clear usr setting
	send_mpu6050_cmd(MPU6050_ADDRESS, 0x64, 0x01); //override register
	send_mpu6050_cmd(MPU6050_ADDRESS, 0x6A, 0x20); //enable master i2c mode
	send_mpu6050_cmd(MPU6050_ADDRESS, 0x34, 0x13); //disable slv4

}

uint8_t mpu6050_init(void)
{
	uint8_t err;
	uint8_t id = 0;
	uint8_t config = 0;


	while(1) {
		err = read_reg(MPU6050_ADDRESS, 0x75, &id);
		if (err)
			continue;
		if (id == 0x68)
			break;
	}



	//Sets sample rate to 1000/1+1 = 500Hz
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_SMPLRT_DIV, 0x01);
	//Disable FSync, 48Hz DLPF
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_CONFIG, 0x03);
	//Disable gyro self tests, scale of 500 degrees/s
	//send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_GYRO_CONFIG, 0b00001000);

	//Disable gyro self tests, scale of 2000 degrees/s
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_GYRO_CONFIG, 0b00011000);

	//Disable accel self tests, scale of +-4g, no DHPF
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_ACCEL_CONFIG, 0b00001000);
	//Freefall threshold of <|0mg|
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_FF_THR, 0x00);
	//Freefall duration limit of 0
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_FF_DUR, 0x00);
	//Motion threshold of >0mg
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_MOT_THR, 0x00);
	//Motion duration of >0s
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_MOT_DUR, 0x00);
	//Zero motion threshold
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_ZRMOT_THR, 0x00);
	//Zero motion duration threshold
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_ZRMOT_DUR, 0x00);
	//Disable sensor output to FIFO buffer
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_FIFO_EN, 0x00);

	//AUX I2C setup
	//Sets AUX I2C to single master control, plus other config
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_I2C_MST_CTRL, 0x00);
	//Setup AUX I2C slaves
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV0_ADDR, 0x00);
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV0_REG, 0x00);
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV0_CTRL, 0x00);
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV1_ADDR, 0x00);
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV1_REG, 0x00);
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV1_CTRL, 0x00);
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV2_ADDR, 0x00);
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV2_REG, 0x00);
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV2_CTRL, 0x00);
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV3_ADDR, 0x00);
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV3_REG, 0x00);
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV3_CTRL, 0x00);
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV4_ADDR, 0x00);
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV4_REG, 0x00);
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV4_DO, 0x00);
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV4_CTRL, 0x00);
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV4_DI, 0x00);


	//MPU6050_RA_I2C_MST_STATUS //Read-only
	//Setup INT pin and AUX I2C pass through
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_INT_PIN_CFG, 0x00);
	//Enable data ready interrupt
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_INT_ENABLE, 0x00);

	//MPU6050_RA_DMP_INT_STATUS                //Read-only
	//MPU6050_RA_INT_STATUS 3A                //Read-only
	//MPU6050_RA_ACCEL_XOUT_H                 //Read-only
	//MPU6050_RA_ACCEL_XOUT_L                 //Read-only
	//MPU6050_RA_ACCEL_YOUT_H                 //Read-only
	//MPU6050_RA_ACCEL_YOUT_L                 //Read-only
	//MPU6050_RA_ACCEL_ZOUT_H                 //Read-only
	//MPU6050_RA_ACCEL_ZOUT_L                 //Read-only
	//MPU6050_RA_TEMP_OUT_H                 //Read-only
	//MPU6050_RA_TEMP_OUT_L                 //Read-only
	//MPU6050_RA_GYRO_XOUT_H                 //Read-only
	//MPU6050_RA_GYRO_XOUT_L                 //Read-only
	//MPU6050_RA_GYRO_YOUT_H                 //Read-only
	//MPU6050_RA_GYRO_YOUT_L                 //Read-only
	//MPU6050_RA_GYRO_ZOUT_H                 //Read-only
	//MPU6050_RA_GYRO_ZOUT_L                 //Read-only
	//MPU6050_RA_EXT_SENS_DATA_00         //Read-only
	//MPU6050_RA_EXT_SENS_DATA_01         //Read-only
	//MPU6050_RA_EXT_SENS_DATA_02         //Read-only
	//MPU6050_RA_EXT_SENS_DATA_03         //Read-only
	//MPU6050_RA_EXT_SENS_DATA_04         //Read-only
	//MPU6050_RA_EXT_SENS_DATA_05         //Read-only
	//MPU6050_RA_EXT_SENS_DATA_06         //Read-only
	//MPU6050_RA_EXT_SENS_DATA_07         //Read-only
	//MPU6050_RA_EXT_SENS_DATA_08         //Read-only
	//MPU6050_RA_EXT_SENS_DATA_09         //Read-only
	//MPU6050_RA_EXT_SENS_DATA_10         //Read-only
	//MPU6050_RA_EXT_SENS_DATA_11         //Read-only
	//MPU6050_RA_EXT_SENS_DATA_12         //Read-only
	//MPU6050_RA_EXT_SENS_DATA_13         //Read-only
	//MPU6050_RA_EXT_SENS_DATA_14         //Read-only
	//MPU6050_RA_EXT_SENS_DATA_15         //Read-only
	//MPU6050_RA_EXT_SENS_DATA_16         //Read-only
	//MPU6050_RA_EXT_SENS_DATA_17         //Read-only
	//MPU6050_RA_EXT_SENS_DATA_18         //Read-only
	//MPU6050_RA_EXT_SENS_DATA_19         //Read-only
	//MPU6050_RA_EXT_SENS_DATA_20         //Read-only
	//MPU6050_RA_EXT_SENS_DATA_21         //Read-only
	//MPU6050_RA_EXT_SENS_DATA_22         //Read-only
	//MPU6050_RA_EXT_SENS_DATA_23         //Read-only
	//MPU6050_RA_MOT_DETECT_STATUS         //Read-only

	//Slave out, dont care
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV0_DO, 0x00);
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV1_DO, 0x00);
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV2_DO, 0x00);
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV3_DO, 0x00);
	//More slave config
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_I2C_MST_DELAY_CTRL, 0x00);
	//Reset sensor signal paths
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_SIGNAL_PATH_RESET, 0x00);
	//Motion detection control
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_MOT_DETECT_CTRL, 0x00);
	//Disables FIFO, AUX I2C, FIFO and I2C reset bits to 0
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_USER_CTRL, 0x00);
	//Sets clock source to gyro reference w/ PLL
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_PWR_MGMT_1, 0b00000010);
	//Controls frequency of wakeups in accel low power mode plus the sensor standby modes
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_PWR_MGMT_2, 0x00);
	//MPU6050_RA_BANK_SEL                        //Not in datasheet
	//MPU6050_RA_MEM_START_ADDR                //Not in datasheet
	//MPU6050_RA_MEM_R_W                        //Not in datasheet
	//MPU6050_RA_DMP_CFG_1                        //Not in datasheet
	//MPU6050_RA_DMP_CFG_2                        //Not in datasheet
	//MPU6050_RA_FIFO_COUNTH                //Read-only
	//MPU6050_RA_FIFO_COUNTL                //Read-only
	//Data transfer to and from the FIFO buffer
	send_mpu6050_cmd(MPU6050_ADDRESS, MPU6050_RA_FIFO_R_W, 0x00);
	//MPU6050_RA_WHO_AM_I                         //Read-only, I2C address

	setup_mpu9150_magneto();


	printf("MPU6050 Setup Complete\r\n");
	mpu6050_compute_drift();
	printf("MPU6050 drift computed\r\n");

}
